@@ -15,15 +15,15 @@
 (defvar *current-internal-error-context*)
 
 (defmacro with-pinned-context-code-object
-      ((&optional (context '*current-internal-error-context*))
-       &body body)
-    (declare (ignorable context))
-    #+(or x86 x86-64)
-    `(progn ,@body)
-    #-(or x86 x86-64)
-    `(with-pinned-objects ((with-code-pages-pinned (:dynamic)
-                             (sb-di::code-object-from-context ,context)))
-       ,@body))
+    ((&optional (context '*current-internal-error-context*))
+     &body body)
+  (declare (ignorable context))
+  #+(or x86 x86-64 arm64)
+  `(progn ,@body)
+  #-(or x86 x86-64 arm64)
+  `(with-pinned-objects ((with-code-pages-pinned (:dynamic)
+                           (sb-di::code-object-from-context ,context)))
+     ,@body))
 
 ;;;; OS-CONTEXT-T
 
@@ -187,6 +187,10 @@
         do (setf (%symbol-function symbol) fun))
   (setf *previous-cpu-routines* nil))
 
+;;; Unless using an arena there is really no way to get a number
+;;; allocated off the heap
+#-x86-64 (defun copy-number-to-heap (n) n)
+
 (defun hexdump (thing &optional (n-words nil wordsp)
                             ;; pass NIL explicitly if T crashes on you
                         (decode t))
