@@ -180,16 +180,13 @@
      (pop args)))
 
 (defmacro def-complex-format-interpreter (char lambda-list &body body)
+  (aver (not (lower-case-p char)))
   (let ((defun-name
-            (intern (format nil
-                            "~:@(~:C~)-FORMAT-DIRECTIVE-INTERPRETER"
-                            char)))
+            (intern (format nil "~:C-INTERPRETER" char)))
         (directive '.directive) ; expose this var to the lambda. it's easiest
-        (directives (if lambda-list (car (last lambda-list)) (sb-xc:gensym "DIRECTIVES"))))
-    `(!cold-init-forms
-      (setf
-       (aref *format-directive-interpreters* ,(char-code (char-upcase char)))
-       (named-lambda ,defun-name (stream ,directive ,directives orig-args args)
+        (directives (if lambda-list (car (last lambda-list)) (gensym "DIRECTIVES"))))
+    `(progn
+       (defun ,defun-name (stream ,directive ,directives orig-args args)
          (declare (ignorable stream orig-args args))
          ,@(if lambda-list
                `((let ,(mapcar (lambda (var)
@@ -198,10 +195,13 @@
                                (butlast lambda-list))
                    (values (progn ,@body) args)))
                `((declare (ignore ,directive ,directives))
-                 ,@body)))))))
+                 ,@body)))
+       (!cold-init-forms
+        (setf (aref *format-directive-interpreters* (char-code ,char))
+              #',defun-name)))))
 
 (defmacro def-format-interpreter (char lambda-list &body body)
-  (let ((directives (sb-xc:gensym "DIRECTIVES")))
+  (let ((directives (gensym "DIRECTIVES")))
     `(def-complex-format-interpreter ,char (,@lambda-list ,directives)
        ,@body
        ,directives)))
