@@ -560,7 +560,8 @@
          (tmp-obj
            (concatenate 'string obj
                         (if *compile-for-effect-only* "-scratch" "-tmp")))
-         (compile-file (ecase mode
+         (compilation-fn
+                       (ecase mode
                          (:host-compile
                           #+abcl ; ABCL complains about its own deficiency and then returns T
                           ;; for warnings and failure. "Unable to compile function" is not our problem,
@@ -590,7 +591,7 @@
                                               *target-compile-file*))))
          (trace-file (if (find :trace-file flags) t nil))
          (block-compile (if (find :block-compile flags) t :specified)))
-    (declare (type function compile-file))
+    (declare (type function compilation-fn))
 
     (ensure-directories-exist obj :verbose cl:*compile-print*) ; host's value
 
@@ -637,10 +638,10 @@
            (report-continue-restart (stream)
              (format stream "Continue, using possibly bogus file ~S" obj)))
       (tagbody
-       retry-compile-file
+       retry-compile
          (multiple-value-bind (output-truename warnings-p failure-p)
              (restart-case
-                 (apply compile-file src
+                 (apply compilation-fn src
                         :output-file tmp-obj
                         :block-compile (and
                                         ;; Block compilation was
@@ -663,7 +664,7 @@
                           '(:trace-file t :print t)))
                (recompile ()
                  :report report-recompile-restart
-                 (go retry-compile-file)))
+                 (go retry-compile)))
            (declare (ignore warnings-p))
            (cond ((not output-truename)
                   (error "couldn't compile ~S" src))
@@ -674,7 +675,7 @@
                                   obj)
                          (recompile ()
                            :report report-recompile-restart
-                           (go retry-compile-file))
+                           (go retry-compile))
                          (continue ()
                            :report report-continue-restart
                            (setf failure-p nil)))
@@ -745,13 +746,24 @@
 ;;; Run the cross-compiler on a file in the source directory tree to
 ;;; produce a corresponding file in the target object directory tree.
 (defun target-compile-stem (stem flags)
-  (funcall *in-target-compilation-mode-fn*
+  (let ((system-tlab-p
+         (or (search "src/pcl" stem)
+             (search "src/code/alieneval" stem)
+             (search "src/code/arena" stem)
+             (search "src/code/debug" stem)
+             (search "src/code/early-defmethod" stem)
+             (search "src/code/format" stem)
+             (search "src/code/brothertree" stem)
+             (search "src/code/avltree" stem))))
+    (funcall *in-target-compilation-mode-fn*
            (lambda ()
-             (progv (list (intern "*SOURCE-NAMESTRING*" "SB-C"))
-                    (list (lpnify-stem stem))
+             (progv (list (intern "*SOURCE-NAMESTRING*" "SB-C")
+                          (intern "*FORCE-SYSTEM-TLAB*" "SB-C"))
+                    (list (lpnify-stem stem)
+                          system-tlab-p)
                (loop
                 (with-simple-restart (recompile "Recompile")
-                  (return (compile-stem stem flags :target-compile))))))))
+                  (return (compile-stem stem flags :target-compile)))))))))
 (compile 'target-compile-stem)
 
 ;;; (This function is not used by the build process, but is intended
