@@ -211,6 +211,13 @@
 (define-binop logior 2 orr)
 (define-binop logxor 2 eor)
 
+(define-vop (fast--/unsigned=>signed fast-unsigned-binop)
+  (:translate -)
+  (:results (r :scs (signed-reg)))
+  (:result-types signed-num)
+  (:generator 5
+    (inst sub r x y)))
+
 (define-binop logandc1 2 bic :swap t
   :constant-test bic-encode-immediate
   :constant-fixnum-test bic-fixnum-encode-immediate
@@ -434,10 +441,10 @@
               ;; Only the first 6 bits count for shifts.
               ;; This sets all bits to 1 if AMOUNT is larger than 63,
               ;; cutting the amount to 63.
-              (inst csinv temp temp zr-tn :lt)
+              (inst csinv temp temp zr-tn :lo)
               (inst asr result number temp))
              (:unsigned
-              (inst csel result number zr-tn :lt)
+              (inst csel result number zr-tn :lo)
               (inst lsr result result temp)))))
 
     (inst b END)
@@ -447,7 +454,7 @@
            (inst lsl result number temp))
           (t
            (inst cmp temp n-word-bits)
-           (inst csel result number zr-tn :lt)
+           (inst csel result number zr-tn :lo)
            (inst lsl result result temp)))
     END))
 
@@ -478,10 +485,10 @@
               ;; Only the first 6 bits count for shifts.
               ;; This sets all bits to 1 if AMOUNT is larger than 63,
               ;; cutting the amount to 63.
-              (inst csinv temp temp zr-tn :lt)
+              (inst csinv temp temp zr-tn :lo)
               (inst asr temp-result number temp))
              (unsigned-reg
-              (inst csel temp-result number zr-tn :lt)
+              (inst csel temp-result number zr-tn :lo)
               (inst lsr temp-result temp-result temp)))))
 
     (inst b END)
@@ -491,7 +498,7 @@
            (inst lsl temp-result number temp))
           (t
            (inst cmp temp n-word-bits)
-           (inst csel temp-result number zr-tn :lt)
+           (inst csel temp-result number zr-tn :lo)
            (inst lsl temp-result temp-result temp)))
     END
     (inst lsl result temp-result n-fixnum-tag-bits)))
@@ -534,10 +541,10 @@
                                                 (specifier-type `(mod ,n-word-bits)))))
                            (inst cmp amount n-word-bits)
                            (cond ((location= amount result)
-                                  (inst csel tmp-tn number zr-tn :lt)
+                                  (inst csel tmp-tn number zr-tn :lo)
                                   (inst lsl result tmp-tn amount))
                                  (t
-                                  (inst csel result number zr-tn :lt)
+                                  (inst csel result number zr-tn :lo)
                                   (inst lsl result result amount))))
                           (t
                            (inst lsl result number amount)))))
@@ -927,7 +934,7 @@
   (:arg-types unsigned-num unsigned-num)
   (:note "inline (unsigned-byte 64) comparison"))
 
-(defmacro define-conditional-vop (tran signed unsigned)
+(defmacro define-conditional-vop (tran signed unsigned &optional addend addend-signed addend-unsigned)
   `(progn
      ,@(loop for (suffix cost signed-p) in
              '((/fixnum 4 t)
@@ -947,14 +954,31 @@
        (:arg-types (:or tagged-num signed-num unsigned-num)
                    (:constant (or signed-word word)))
        (:info y)
+       (:vop-var vop)
        (:policy :fast-safe)
        ,(if (eq signed unsigned)
             `(:conditional ,signed)
             `(:conditional
               :after-sc-selection
-              (if (sc-is x unsigned-reg)
-                  ,unsigned
-                  ,signed)))
+              (flet ((try (y)
+                       (let ((y (if (sc-is x any-reg)
+                                    (fixnumize y)
+                                    y)))
+                         (flet ((try (constant)
+                                  (add-sub-immediate-p constant)))
+                           (or (try y)
+                               (try (ldb (byte 64 0) (- y))))))))
+                (cond ((and (not (try y))
+                            (try (+ y ,addend)))
+                       (setf (car (vop-codegen-info vop))
+                             (+ y ,addend))
+                       (if (sc-is x unsigned-reg)
+                           ,addend-unsigned
+                           ,addend-signed))
+                      (t
+                       (if (sc-is x unsigned-reg)
+                           ,unsigned
+                           ,signed))))))
        (:generator 2
          (let ((y (if (sc-is x any-reg)
                       (fixnumize y)
@@ -969,10 +993,64 @@
                  (try (ldb (byte 64 0) (- y)) t)
                  (inst cmp x (load-immediate-word tmp-tn y)))))))))
 
-(define-conditional-vop < :lt :lo)
-(define-conditional-vop > :gt :hi)
+(define-conditional-vop < :lt :lo -1 :le :ls)
+(define-conditional-vop > :gt :hi 1 :ge :hs)
 (define-conditional-vop eql :eq :eq)
 
+(define-vop (<-unsigned-signed)
+  (:translate <)
+  (:args (unsigned :scs (unsigned-reg))
+         (signed :scs (signed-reg)))
+  (:arg-types unsigned-num signed-num)
+  (:conditional :lo)
+  (:policy :fast-safe)
+  (:generator 7
+    (inst cmp signed 0)
+    (inst ccmp unsigned signed :ge #b0010)))
+
+(define-vop (>-unsigned-signed <-unsigned-signed)
+  (:translate >)
+  (:conditional :hi))
+
+(define-vop (<-signed-unsigned)
+  (:translate <)
+  (:args (signed :scs (signed-reg))
+         (unsigned :scs (unsigned-reg)))
+  (:arg-types signed-num unsigned-num)
+  (:conditional :lo)
+  (:policy :fast-safe)
+  (:generator 7
+    (inst cmp signed 0)
+    (inst ccmp signed unsigned :ge #b0000)))
+
+(define-vop (>-signed-unsigned <-signed-unsigned)
+  (:translate >)
+  (:conditional :hi))
+
+(define-vop (eql-unsigned-signed)
+  (:translate eql)
+  (:args (unsigned :scs (unsigned-reg))
+         (signed :scs (signed-reg)))
+  (:arg-types unsigned-num signed-num)
+  (:conditional :eq)
+  (:policy :fast-safe)
+  (:generator 7
+    (inst cmp signed 0)
+    (inst ccmp unsigned signed :ge #b0000)))
+
+(define-vop (eql-signed-unsigned eql-unsigned-signed)
+  (:args (signed :scs (signed-reg))
+         (unsigned :scs (unsigned-reg)))
+  (:arg-types signed-num unsigned-num))
+
+(define-vop (eq-unsigned-signed eql-unsigned-signed)
+  (:variant-cost 6)
+  (:translate eq))
+
+(define-vop (eq-signed-unsigned eql-signed-unsigned)
+  (:variant-cost 6)
+  (:translate eq))
+
 (define-vop (generic-eql/fixnum fast-if-eql/fixnum)
   (:args (x :scs (any-reg descriptor-reg))
          (y :scs (any-reg)))
@@ -1006,12 +1084,27 @@
   (:policy :fast-safe)
   (:conditional :ne)
   (:generator 2
-    (let ((y (if (sc-is x any-reg)
-                 (fixnumize y)
-                 y)))
-      (if (encode-logical-immediate y)
-          (inst tst x y)
-          (inst tst x (load-immediate-word tmp-tn y))))))
+    (block nil
+      (if (= y most-positive-word)
+          (inst cmp x 0)
+          (let ((y (if (sc-is x any-reg)
+                       (cond ((fixnump y)
+                              (fixnumize y))
+                             ((let ((y (ldb (byte 64 0) y)))
+                                ;; Only a negative fixnum will match that bit,
+                                ;; so if the next bit is also 1 it can be shifted left.
+                                (when (and (= (ldb (byte 1 62) y) 1)
+                                           (encode-logical-immediate
+                                            (ldb (byte 64 0) (ash y 1))))
+                                  (return (inst tst x (ldb (byte 64 0) (ash y 1)))))))
+                             (t
+                              (return
+                                (inst tst (load-immediate-word tmp-tn y)
+                                      (asr x n-fixnum-tag-bits)))))
+                       y)))
+            (if (encode-logical-immediate y)
+                (inst tst x y)
+                (inst tst x (load-immediate-word tmp-tn y))))))))
 
 (define-source-transform lognand (x y)
   `(lognot (logand ,x ,y)))
@@ -1118,6 +1211,17 @@
 (define-full-setter bignum-set * bignum-digits-offset other-pointer-lowtag
   (unsigned-reg) unsigned-num sb-bignum:%bignum-set)
 
+(define-partial-reffer sb-c::%half-bignum-ref bignum
+  :word nil bignum-digits-offset other-pointer-lowtag (unsigned-reg signed-reg)
+  positive-fixnum
+  sb-bignum:%half-bignum-ref)
+
+(define-partial-setter sb-c::%half-bignum-set bignum
+  :word bignum-digits-offset other-pointer-lowtag (unsigned-reg signed-reg)
+  positive-fixnum
+  (setf sb-bignum:%half-bignum-ref))
+
+
 (define-vop (digit-0-or-plus)
   (:translate sb-bignum:%digit-0-or-plusp)
   (:policy :fast-safe)
@@ -1126,8 +1230,8 @@
   (:conditional)
   (:info target not-p)
   (:generator 2
-     (inst cmp digit 0)
-     (inst b (if not-p :lt :ge) target)))
+    (inst cmp digit 0)
+    (inst b (if not-p :lt :ge) target)))
 
 (define-vop (add-w/carry)
   (:translate sb-bignum:%add-with-carry)
@@ -1161,7 +1265,7 @@
     (inst cmp c 1) ;; Set carry if (fixnum 0 or 1) c=0, else clear.
     (inst sbcs result a b)
     (unless (eq (tn-kind borrow) :unused)
-     (inst cset borrow :cs))))
+      (inst cset borrow :cs))))
 
 (define-vop (bignum-mult-and-add-3-arg)
   (:translate sb-bignum:%multiply-and-add)
@@ -1272,6 +1376,18 @@
         (unless (= i 64)
           (inst adc rem rem rem))))))
 
+(define-vop (half-bignum-floor bignum-floor)
+  (:translate sb-bignum:%half-bigfloor)
+  (:args (div-high :scs (unsigned-reg) :to :save)
+         (div-low :scs (unsigned-reg) :target x)
+         (divisor :scs (unsigned-reg) :to (:result 1)))
+  (:temporary (:sc unsigned-reg :from (:argument 1)) x)
+  (:generator 30
+    (move x div-low)
+    (inst bfm x div-high 32 31)
+    (inst udiv quo x divisor)
+    (inst msub rem quo divisor x)))
+
 (define-vop (signify-digit)
   (:translate sb-bignum:%fixnum-digit-with-correct-sign)
   (:policy :fast-safe)
@@ -1304,3 +1420,495 @@
   (:translate sb-bignum:%ashl)
   (:generator 1
     (inst lsl result digit count)))
+
+(define-vop ()
+  (:translate sb-c::fixnum*)
+  (:args (x :scs (any-reg))
+         (y :scs (signed-reg immediate)))
+  (:arg-types tagged-num tagged-num (:constant t))
+  (:info type)
+  (:temporary (:sc signed-reg) high)
+  (:results (r :scs (any-reg) :from :load))
+  (:result-types tagged-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::mul-overflow-error r high)))
+      (let (value)
+        (when (sc-is y immediate)
+          (load-immediate-word high (setf value (tn-value y)))
+          (setf y high))
+        (if (and value
+                 (plusp value)
+                 (= (logcount value) 1))
+            (inst lsl r x (1- (integer-length value)))
+            (inst mul r x y))
+        (inst smulh high x y))
+      (inst cmp high (asr r 63))
+      (inst b :ne error))))
+
+(define-vop ()
+  (:translate sb-c::signed*)
+  (:args (x :scs (signed-reg))
+         (y :scs (signed-reg immediate)))
+  (:arg-types signed-num signed-num (:constant t))
+  (:info type)
+  (:temporary (:sc signed-reg) high)
+  (:results (r :scs (signed-reg) :from :load))
+  (:result-types signed-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::mul-overflow-error r high)))
+      (let (value)
+        (when (sc-is y immediate)
+          (load-immediate-word high (setf value (tn-value y)))
+          (setf y high))
+        (if (and value
+                 (plusp value)
+                 (= (logcount value) 1))
+            (inst lsl r x (1- (integer-length value)))
+            (inst mul r x y))
+        (inst smulh high x y))
+      (inst cmp high (asr r 63))
+      (inst b :ne error))))
+
+(define-vop ()
+  (:translate sb-c::unsigned*)
+  (:args (x :scs (unsigned-reg))
+         (y :scs (unsigned-reg immediate)))
+  (:arg-types unsigned-num unsigned-num (:constant t))
+  (:info type)
+  (:temporary (:sc unsigned-reg) high)
+  (:results (r :scs (unsigned-reg) :from :load))
+  (:result-types unsigned-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::mul-overflow-error r high)))
+      (let ((value (and (sc-is y immediate)
+                        (tn-value y))))
+
+        (cond ((and value
+                    (plusp value)
+                    (= (logcount value) 1))
+               (let ((shift (1- (integer-length value))))
+                 (inst lsl r x shift)
+                 (inst lsr high x (- n-word-bits shift))))
+              (t
+               (when (sc-is y immediate)
+                 (load-immediate-word high value)
+                 (setf y high))
+               (inst mul r x y)
+               (inst umulh high x y))))
+      (inst cbnz high error))))
+
+(define-vop (*/signed=>integer)
+  (:translate *)
+  (:args (x :scs (signed-reg))
+         (y :scs (signed-reg immediate)))
+  (:arg-types signed-num signed-num)
+  (:temporary (:sc signed-reg) high low)
+  (:temporary (:sc signed-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
+  (:results (r :scs (descriptor-reg)))
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 10
+    (let (value)
+      (when (sc-is y immediate)
+        (load-immediate-word high (setf value (tn-value y)))
+        (setf y high))
+      (if (and value
+               (plusp value)
+               (= (logcount value) 1))
+          (inst lsl low x (1- (integer-length value)))
+          (inst mul low x y))
+      (inst smulh high x y))
+    (inst mov header (bignum-header-for-length 2))
+    (inst cmp high (asr low 63))
+    (inst b :ne allocate)
+    (inst adds r low low)
+    (inst b :vc done)
+    (inst mov header (bignum-header-for-length 1))
+    #+bignum-assertions
+    (inst mov high 0)
+    allocate
+    (with-fixed-allocation
+        (r lr nil (+ 2 bignum-digits-offset))
+      (storew-pair header 0 low bignum-digits-offset tmp-tn)
+      (storew high tmp-tn 2))
+    DONE))
+
+(define-vop (*/unsigned=>integer)
+  (:translate *)
+  (:args (x :scs (unsigned-reg))
+         (y :scs (unsigned-reg immediate)))
+  (:arg-types unsigned-num unsigned-num)
+  (:temporary (:sc unsigned-reg) high low)
+  (:temporary (:sc unsigned-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
+  (:results (r :scs (descriptor-reg)))
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 12
+    (let ((value (and (sc-is y immediate)
+                        (tn-value y))))
+        (cond ((and value
+                    (plusp value)
+                    (= (logcount value) 1))
+               (let ((shift (1- (integer-length value))))
+                 (inst lsl low x shift)
+                 (inst lsr high x (- n-word-bits shift))))
+              (t
+               (when (sc-is y immediate)
+                 (load-immediate-word high value)
+                 (setf y high))
+               (inst mul low x y)
+               (inst umulh high x y))))
+    (inst mov header (bignum-header-for-length 3))
+    (inst tbnz high 63 allocate)
+    (inst mov header (bignum-header-for-length 2))
+    (inst cbnz high allocate)
+    (inst tbnz low 63 allocate)
+    (inst adds r low low)
+    (inst b :vc done)
+    (inst mov header (bignum-header-for-length 1))
+    allocate
+    (with-fixed-allocation
+        (r lr nil (+ 2 bignum-digits-offset))
+      (storew-pair header 0 low bignum-digits-offset tmp-tn)
+      (storew high tmp-tn 2))
+    DONE))
+
+(define-vop (+/signed=>integer)
+  (:translate +)
+  (:args (x :scs (signed-reg))
+         (y :scs (signed-reg immediate)))
+  (:arg-types signed-num signed-num)
+  (:temporary (:sc signed-reg) high low)
+  (:temporary (:sc signed-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
+  (:results (r :scs (descriptor-reg)))
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 10
+    (if (sc-is y immediate)
+        (let ((y (tn-value y)))
+          (if (minusp y)
+              (inst subs low x (add-sub-immediate (- y) high))
+              (inst adds low x (add-sub-immediate y high))))
+        (inst adds low x y))
+    (inst csetm high :cs)
+    (inst mov header (bignum-header-for-length 2))
+    (inst b :vs allocate)
+    (inst adds r low low)
+    (inst b :vc done)
+    (inst mov header (bignum-header-for-length 1))
+    allocate
+    (with-fixed-allocation
+        (r lr nil (+ 2 bignum-digits-offset))
+      (storew-pair header 0 low bignum-digits-offset tmp-tn)
+      (storew high tmp-tn 2))
+    DONE))
+
+(define-vop (-/signed=>integer)
+  (:translate -)
+  (:args (x :scs (signed-reg))
+         (y :scs (signed-reg immediate)))
+  (:arg-types signed-num signed-num)
+  (:temporary (:sc signed-reg) high low)
+  (:temporary (:sc signed-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
+  (:results (r :scs (descriptor-reg)))
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 10
+    (if (sc-is y immediate)
+        (let ((y (tn-value y)))
+          (if (minusp y)
+              (inst adds low x (add-sub-immediate (- y) high))
+              (inst subs low x (add-sub-immediate y high))))
+        (inst subs low x y))
+    (inst csetm high :cs)
+    (inst mov header (bignum-header-for-length 2))
+    (inst b :vs allocate)
+    (inst adds r low low)
+    (inst b :vc done)
+    (inst mov header (bignum-header-for-length 1))
+    allocate
+    (with-fixed-allocation
+        (r lr nil (+ 2 bignum-digits-offset))
+      (storew-pair header 0 low bignum-digits-offset tmp-tn)
+      (storew high tmp-tn 2))
+    DONE))
+
+(define-vop (+/unsigned=>integer)
+  (:translate +)
+  (:args (x :scs (unsigned-reg))
+         (y :scs (unsigned-reg immediate)))
+  (:arg-types unsigned-num unsigned-num)
+  (:temporary (:sc unsigned-reg) high low)
+  (:temporary (:sc unsigned-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
+  (:results (r :scs (descriptor-reg)))
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 10
+    (inst adds low x (if (sc-is y immediate)
+                         (add-sub-immediate (tn-value y) high)
+                         y))
+    (inst mov header (bignum-header-for-length 2))
+    (inst cset high :cs)
+    (inst b :cs allocate)
+    (inst b :mi allocate)
+    (inst adds r low low)
+    (inst b :vc done)
+    (inst mov header (bignum-header-for-length 1))
+    allocate
+    (with-fixed-allocation
+        (r lr nil (+ 2 bignum-digits-offset))
+      (storew-pair header 0 low bignum-digits-offset tmp-tn)
+      (storew high tmp-tn 2))
+    DONE))
+
+(define-vop (-/unsigned=>integer)
+  (:translate -)
+  (:args (x :scs (unsigned-reg))
+         (y :scs (unsigned-reg immediate)))
+  (:arg-types unsigned-num unsigned-num)
+  (:temporary (:sc unsigned-reg) high low)
+  (:temporary (:sc unsigned-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
+  (:results (r :scs (descriptor-reg)))
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 10
+    (inst subs low x (if (sc-is y immediate)
+                         (add-sub-immediate (tn-value y) high)
+                         y))
+    (inst mov header (bignum-header-for-length 2))
+    (inst csetm high :cc)
+    (inst b :cc negative)
+    (inst b :mi allocate)
+    (inst b positive)
+    negative
+    (inst b :pl allocate)
+    positive
+    (inst adds r low low)
+    (inst b :vc done)
+    (inst mov header (bignum-header-for-length 1))
+    allocate
+    (with-fixed-allocation
+        (r lr nil (+ 2 bignum-digits-offset))
+      (storew-pair header 0 low bignum-digits-offset tmp-tn)
+      (storew high tmp-tn 2))
+    DONE))
+
+(define-vop ()
+  (:translate sb-c::signed+)
+  (:args (x :scs (signed-reg))
+         (y :scs (signed-reg immediate)))
+  (:arg-types signed-num signed-num (:constant t))
+  (:info type)
+  (:results (r :scs (signed-reg)))
+  (:result-types signed-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::add-sub-overflow-error r)))
+      (if (sc-is y immediate)
+          (let ((y (tn-value y)))
+            (if (minusp y)
+                (inst subs r x (add-sub-immediate (- y)))
+                (inst adds r x (add-sub-immediate y))))
+          (inst adds r x y))
+      (inst b :vs error))))
+
+(define-vop ()
+  (:translate sb-c::signed-)
+  (:args (x :scs (signed-reg))
+         (y :scs (signed-reg immediate)))
+  (:arg-types signed-num signed-num (:constant t))
+  (:info type)
+  (:results (r :scs (signed-reg)))
+  (:result-types signed-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::add-sub-overflow-error r)))
+      (if (sc-is y immediate)
+          (let ((y (tn-value y)))
+            (if (minusp y)
+                (inst adds r x (add-sub-immediate (- y)))
+                (inst subs r x (add-sub-immediate y))))
+          (inst subs r x y))
+      (inst b :vs error))))
+
+(define-vop ()
+  (:translate sb-c::unsigned+)
+  (:args (x :scs (unsigned-reg))
+         (y :scs (unsigned-reg immediate)))
+  (:arg-types unsigned-num unsigned-num (:constant t))
+  (:info type)
+  (:results (r :scs (unsigned-reg)))
+  (:result-types unsigned-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::add-sub-overflow-error r)))
+      (when (sc-is y immediate)
+        (setf y (add-sub-immediate (tn-value y))))
+      (inst adds r x y)
+      (inst b :cs error))))
+
+(define-vop ()
+  (:translate sb-c::unsigned-)
+  (:args (x :scs (unsigned-reg))
+         (y :scs (unsigned-reg immediate)))
+  (:arg-types unsigned-num unsigned-num (:constant t))
+  (:info type)
+  (:results (r :scs (unsigned-reg)))
+  (:result-types unsigned-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::add-sub-overflow-error r)))
+      (when (sc-is y immediate)
+        (setf y (add-sub-immediate (tn-value y))))
+      (inst subs r x y)
+      (inst b :cc error))))
+
+(define-vop ()
+  (:translate sb-c::unsigned+signed)
+  (:args (x :scs (unsigned-reg) :to :result)
+         (y :scs (signed-reg immediate)))
+  (:arg-types unsigned-num signed-num (:constant t))
+  (:info type)
+  (:results (r :scs (unsigned-reg)))
+  (:temporary (:scs (unsigned-reg) :from (:argument 2)) temp)
+  (:result-types unsigned-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::add-sub-overflow-error r)))
+      (if (sc-is y immediate)
+          (assemble ()
+            (let ((value (tn-value y)))
+              (setf y (add-sub-immediate (abs value)))
+              (cond ((plusp value)
+                     (inst adds r x y)
+                     (inst b :cs error))
+                    (t
+                     (inst subs r x y)
+                     (inst b :cc error)))))
+          (assemble ()
+            (when (sc-is y immediate)
+              (setf y (add-sub-immediate (tn-value y))))
+            (inst tbnz y 63 neg)
+            (inst adds r x y)
+            (inst b :cs error)
+            (inst b done)
+            neg
+            (inst neg temp y)
+            (inst subs r x temp)
+            (inst b :cc error)
+            done)))))
+
+(define-vop ()
+  (:translate sb-c::unsigned-signed)
+  (:args (x :scs (unsigned-reg) :to :result)
+         (y :scs (signed-reg immediate)))
+  (:arg-types unsigned-num signed-num (:constant t))
+  (:info type)
+  (:results (r :scs (unsigned-reg)))
+  (:temporary (:scs (unsigned-reg) :from (:argument 2)) temp)
+  (:result-types unsigned-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::add-sub-overflow-error r)))
+      (if (sc-is y immediate)
+          (assemble ()
+            (let ((value (tn-value y)))
+              (setf y (add-sub-immediate (abs value)))
+              (cond ((minusp value)
+                     (inst adds r x y)
+                     (inst b :cs error))
+                    (t
+                     (inst subs r x y)
+                     (inst b :cc error)))))
+          (assemble ()
+            (when (sc-is y immediate)
+              (setf y (add-sub-immediate (tn-value y))))
+            (inst tbz y 63 pos)
+            (inst neg temp y)
+            (inst adds r x temp)
+            (inst b :cs error)
+            (inst b done)
+            pos
+            (inst subs r x y)
+            (inst b :cc error)
+            done)))))
+
+(define-vop (signum-signed signed-unop)
+  (:args (x :scs (signed-reg any-reg) :target res))
+  (:translate signum)
+  (:generator 4
+    (inst cmp x 0)
+    (inst cset tmp-tn :ne)
+    (inst csinv res tmp-tn zr-tn :ge)))
+
+(define-vop (signum-unsigned unsigned-unop)
+  (:args (x :scs (unsigned-reg any-reg) :target res))
+  (:translate signum)
+  (:generator 3
+    (inst cmp x 0)
+    (inst cset res :ne)))
+
+(define-vop ()
+  (:translate fastrem-32)
+  (:policy :fast-safe)
+  (:args (dividend :scs (unsigned-reg))
+         (c :scs (unsigned-reg))
+         (divisor :scs (unsigned-reg)))
+  (:arg-types unsigned-num unsigned-num unsigned-num)
+  (:results (remainder :scs (unsigned-reg)))
+  (:result-types unsigned-num)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 10
+    (inst mul temp dividend c)
+    (inst and temp temp #xFFFFFFFF) ; drop the high 32 bits, keep the low 32 bits
+    (inst mul temp temp divisor)
+    (inst lsr remainder temp 32))) ; take the high 32 bits
+(define-vop ()
+  (:translate fastrem-64)
+  (:policy :fast-safe)
+  (:args (dividend :scs (unsigned-reg))
+         (c :scs (unsigned-reg))
+         (divisor :scs (unsigned-reg)))
+  (:arg-types unsigned-num unsigned-num unsigned-num)
+  (:results (remainder :scs (unsigned-reg)))
+  (:result-types unsigned-num)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 10
+    (inst mul temp dividend c) ; want only the low 64 bits
+    (inst umulh remainder temp divisor))) ; want only the high 64 bits
