@@ -37,50 +37,21 @@
                 (first flags))
             dest)))
 
-(define-load-time-global *cmov-ptype-representation-vop*
-  (mapcan (lambda (entry)
-            (destructuring-bind (ptypes &optional sc vop)
-                entry
-              (mapcar (if (and vop sc)
-                          (lambda (ptype)
-                            (list ptype sc vop))
-                          #'list)
-                      (ensure-list ptypes))))
-          '((t descriptor-reg move-if/descriptor)
-            ((fixnum positive-fixnum) any-reg move-if/descriptor)
-            ((unsigned-byte-64 unsigned-byte-63) unsigned-reg move-if/word)
-            (signed-byte-64 signed-reg move-if/word)
-            (character character-reg move-if/char)
-            ((single-float complex-single-float
-              double-float complex-double-float))
-            (system-area-pointer sap-reg move-if/sap))))
-
-(defun convert-conditional-move-p (node dst-tn x-tn y-tn)
-  (declare (ignore node))
-  (let* ((ptype (sb-c::tn-primitive-type dst-tn))
-         (name  (sb-c:primitive-type-name ptype))
-         (param (cdr (or (assoc name *cmov-ptype-representation-vop*)
-                         '(t descriptor-reg move-if/descriptor)))))
-    (when param
-      (destructuring-bind (representation vop) param
-        (let ((scn (sc-number-or-lose representation)))
-          (labels ((make-tn (tn)
-                     (cond ((and (tn-sc tn)
-                                 (or
-                                  (and
-                                   (sc-is tn immediate)
-                                   (eq (tn-value tn) 0))
-                                  (and
-                                   (sc-is tn descriptor-reg)
-                                   (eql (tn-offset tn) null-offset))))
-                            tn)
-                           (t
-                            (make-representation-tn ptype scn)))))
-            (values vop
-                    (make-tn x-tn) (make-tn y-tn)
-                    (make-tn dst-tn)
-                    nil)))))))
-
+(defun convert-conditional-move-p (dst-tn)
+  (sc-case dst-tn
+    ((descriptor-reg any-reg)
+     'move-if/descriptor)
+    ((unsigned-reg signed-reg)
+     'move-if/word)
+    (double-reg
+     'move-if/double)
+    (single-reg
+     'move-if/single)
+    (sap-reg
+     'move-if/sap)
+    (character-reg
+     'move-if/char)
+    (t)))
 
 (define-vop (move-if)
   (:args (then) (else))
@@ -112,6 +83,30 @@
   (def-move-if move-if/char character (character-reg zero))
   (def-move-if move-if/sap system-area-pointer (sap-reg zero)))
 
+(define-vop (move-if/double move-if)
+  (:args (then :scs (double-reg)) (else :scs (double-reg)))
+  (:arg-types double-float double-float)
+  (:results (res :scs (double-reg)))
+  (:result-types double-float)
+  (:generator 1
+    (let ((not-p (eq (first flags) 'not)))
+      (when not-p (pop flags))
+      (cond ((null (rest flags))
+             (inst fcsel res then else (if not-p
+                                          (negate-condition (car flags))
+                                          (car flags))))
+            (not-p
+             (dolist (flag flags)
+               (inst fcsel res else then flag)))
+            (t
+             (dolist (flag flags)
+               (inst fcsel res then else flag)))))))
+
+(define-vop (move-if/single move-if/double)
+  (:args (then :scs (single-reg)) (else :scs (single-reg)))
+  (:arg-types single-float single-float)
+  (:results (res :scs (single-reg)))
+  (:result-types single-float))
 
 ;;;; Conditional VOPs:
 
@@ -129,7 +124,7 @@
   (:conditional :eq)
   (:policy :fast-safe)
   (:translate eq)
-  (:generator 6
+  (:generator 7
     (let ((value (sc-case y
                    (immediate
                     (let ((value (tn-value y)))
