@@ -116,7 +116,7 @@
 ;;; SOURCE-NAME and DEBUG-NAME. But I (WHN) don't use &AUX bindings,
 ;;; so I'm not motivated. Patches will be accepted...
 (defun ir1-convert-aux-bindings (start next result body aux-vars aux-vals
-                                 post-binding-lexenv)
+                                 post-binding-lexenv &key value-source-forms)
   (declare (type ctran start next) (type (or lvar null) result)
            (list body aux-vars aux-vals))
   (if (null aux-vars)
@@ -131,10 +131,12 @@
                                           :post-binding-lexenv post-binding-lexenv
                                           :debug-name (debug-name
                                                        '&aux-bindings
-                                                       aux-vars))))
+                                                       aux-vars)
+                                          :value-source-forms (rest value-source-forms))))
         (reference-leaf start ctran fun-lvar fun)
         (ir1-convert-combination-args fun-lvar ctran next result
-                                      (list (first aux-vals)))))
+                                      (list (first aux-vals))
+                                      :arg-source-forms (list (first value-source-forms)))))
   (values))
 
 ;;; This is similar to IR1-CONVERT-PROGN-BODY except that code to bind
@@ -149,13 +151,15 @@
 ;;; to start a block outside of this cleanup, causing cleanup code to
 ;;; be emitted when the scope is exited.
 (defun ir1-convert-special-bindings
-    (start next result body aux-vars aux-vals svars post-binding-lexenv)
+    (start next result body aux-vars aux-vals svars post-binding-lexenv
+     &key value-source-forms)
   (declare (type ctran start next) (type (or lvar null) result)
            (list body aux-vars aux-vals svars))
   (cond
    ((null svars)
     (ir1-convert-aux-bindings start next result body aux-vars aux-vals
-                              post-binding-lexenv))
+                              post-binding-lexenv
+                              :value-source-forms value-source-forms))
    (t
     (ctran-starts-block next)
     (let ((cleanup (make-cleanup :kind :special-bind))
@@ -170,7 +174,8 @@
         (ir1-convert-special-bindings cleanup-ctran next result
                                       body aux-vars aux-vals
                                       (rest svars)
-                                      post-binding-lexenv)))))
+                                      post-binding-lexenv
+                                      :value-source-forms value-source-forms)))))
   (values))
 
 ;;; Create a lambda node out of some code, returning the result. The
@@ -199,8 +204,8 @@
                                 debug-name
                                 (note-lexical-bindings t)
                                 post-binding-lexenv
-                                system-lambda
-                                local-policy)
+                                local-policy
+                                value-source-forms)
   (declare (list body vars aux-vars aux-vals))
 
   ;; We're about to try to put new blocks into *CURRENT-COMPONENT*.
@@ -211,7 +216,6 @@
                               :bind bind
                               :%source-name source-name
                               :%debug-name debug-name
-                              :system-lambda-p system-lambda
                               :lexenv (if local-policy
                                           (make-lexenv :policy local-policy)
                                           *lexenv*)))
@@ -269,7 +273,8 @@
             (ir1-convert-special-bindings postbind-ctran result-ctran
                                           result-lvar body
                                           aux-vars aux-vals (svars)
-                                          post-binding-lexenv)))))
+                                          post-binding-lexenv
+                                          :value-source-forms value-source-forms)))))
 
     (link-blocks (component-head *current-component*) (node-block bind))
     (push lambda (component-new-functionals *current-component*))
@@ -331,8 +336,7 @@
                                           :debug-name (debug-name
                                                        '&optional-processor
                                                        name)
-                                          :note-lexical-bindings nil
-                                          :system-lambda t)))))
+                                          :note-lexical-bindings nil)))))
     (mapc (lambda (var arg-var)
             (when (cdr (leaf-refs arg-var))
               (setf (leaf-ever-used var) t)))
@@ -350,8 +354,7 @@
                                         vars supplied-p-p body
                                         aux-vars aux-vals
                                         source-name debug-name
-                                        force post-binding-lexenv
-                                        system-lambda)
+                                        force post-binding-lexenv)
   (declare (type optional-dispatch res)
            (list default-vars default-vals entry-vars entry-vals vars body
                  aux-vars aux-vals))
@@ -371,7 +374,7 @@
                   (list* t arg-name entry-vals)
                   (rest vars) t body aux-vars aux-vals
                   source-name debug-name
-                  force post-binding-lexenv system-lambda)
+                  force post-binding-lexenv)
                  (ir1-convert-hairy-args
                   res
                   (cons arg default-vars)
@@ -380,7 +383,7 @@
                   (cons arg-name entry-vals)
                   (rest vars) supplied-p-p body aux-vars aux-vals
                   source-name debug-name
-                  force post-binding-lexenv system-lambda))))
+                  force post-binding-lexenv))))
 
     ;; We want to delay converting the entry, but there exist
     ;; problems: hidden references should not be established to
@@ -449,7 +452,8 @@
                                  :type (leaf-type var)
                                  :where-from (leaf-where-from var))))
 
-    (let* ((n-context (gensym "N-CONTEXT-"))
+    (let* ((*allow-instrumenting* nil)
+           (n-context (gensym "N-CONTEXT-"))
            (context-temp (make-lambda-var :%source-name n-context
                                           :arg-info (make-arg-info :kind :more-context)))
            (n-count (gensym "N-COUNT-"))
@@ -587,8 +591,7 @@
                                ,@(arg-vals))))
                  (arg-vars)
                  :debug-name (debug-name '&more-processor name)
-                 :note-lexical-bindings nil
-                 :system-lambda t)))
+                 :note-lexical-bindings nil)))
         (setf (optional-dispatch-more-entry res)
               (register-entry-point ep res)))))
 
@@ -614,7 +617,7 @@
 (defun ir1-convert-more (res default-vars default-vals entry-vars entry-vals
                          rest more-context more-count keys supplied-p-p
                          body aux-vars aux-vals source-name debug-name
-                         post-binding-lexenv system-lambda)
+                         post-binding-lexenv)
   (declare (type optional-dispatch res)
            (list default-vars default-vals entry-vars entry-vals keys body
                  aux-vars aux-vals))
@@ -695,8 +698,7 @@
                         :aux-vals (append (bind-vals) aux-vals)
                         :post-binding-lexenv post-binding-lexenv
                         :source-name source-name
-                        :debug-name debug-name
-                        :system-lambda system-lambda))
+                        :debug-name debug-name))
            (name (or debug-name source-name))
            (last-entry (convert-optional-entry main-entry default-vars
                                                (main-vals) () name)))
@@ -751,8 +753,7 @@
                                vars supplied-p-p body aux-vars
                                aux-vals
                                source-name debug-name
-                               force post-binding-lexenv
-                               system-lambda)
+                               force post-binding-lexenv)
   (declare (type optional-dispatch res)
            (list default-vars default-vals entry-vars entry-vals vars body
                  aux-vars aux-vals))
@@ -764,7 +765,7 @@
                                entry-vars entry-vals
                                nil nil nil vars supplied-p-p body aux-vars
                                aux-vals source-name debug-name
-                               post-binding-lexenv system-lambda)
+                               post-binding-lexenv)
              (let* ((name (or debug-name source-name))
                     (fun (ir1-convert-lambda-body
                          body (reverse default-vars)
@@ -772,8 +773,7 @@
                          :aux-vals aux-vals
                          :post-binding-lexenv post-binding-lexenv
                          :source-name source-name
-                         :debug-name debug-name
-                         :system-lambda system-lambda)))
+                         :debug-name debug-name)))
 
                (setf (optional-dispatch-main-entry res) fun)
                (register-entry-point fun res)
@@ -792,7 +792,7 @@
            (ir1-convert-hairy-args res nvars nvals nvars nvals
                                    (rest vars) nil body aux-vars aux-vals
                                    source-name debug-name
-                                   nil post-binding-lexenv system-lambda)))
+                                   nil post-binding-lexenv)))
         (t
          (let* ((arg (first vars))
                 (info (lambda-var-arg-info arg))
@@ -804,8 +804,7 @@
                          entry-vars entry-vals vars supplied-p-p body
                          aux-vars aux-vals
                          source-name debug-name
-                         force post-binding-lexenv
-                         system-lambda)))
+                         force post-binding-lexenv)))
                 ;; See GENERATE-OPTIONAL-DEFAULT-ENTRY.
                 (push (if (lambda-p ep)
                           (register-entry-point
@@ -825,20 +824,20 @@
                                 arg nil nil (rest vars) supplied-p-p body
                                 aux-vars aux-vals
                                 source-name debug-name
-                                post-binding-lexenv system-lambda))
+                                post-binding-lexenv))
              (:more-context
               (ir1-convert-more res default-vars default-vals
                                 entry-vars entry-vals
                                 nil arg (second vars) (cddr vars) supplied-p-p
                                 body aux-vars aux-vals
                                 source-name debug-name
-                                post-binding-lexenv system-lambda))
+                                post-binding-lexenv))
              (:keyword
               (ir1-convert-more res default-vars default-vals
                                 entry-vars entry-vals
                                 nil nil nil vars supplied-p-p body aux-vars
                                 aux-vals source-name debug-name
-                                post-binding-lexenv system-lambda)))))))
+                                post-binding-lexenv)))))))
 
 ;;; This function deals with the case where we have to make an
 ;;; OPTIONAL-DISPATCH to represent a LAMBDA. We cons up the result and
@@ -847,7 +846,7 @@
 (defun ir1-convert-hairy-lambda (body vars keyp allowp aux-vars aux-vals
                                  &key post-binding-lexenv
                                  (source-name '.anonymous.)
-                                 debug-name system-lambda)
+                                 debug-name)
   (declare (list body vars aux-vars aux-vals))
   (aver (or debug-name (neq '.anonymous. source-name)))
   (let ((res (make-optional-dispatch :arglist vars
@@ -859,8 +858,7 @@
         (min (or (position-if #'lambda-var-arg-info vars) (length vars))))
     (aver-live-component *current-component*)
     (ir1-convert-hairy-args res () () () () vars nil body aux-vars aux-vals
-                            source-name debug-name nil post-binding-lexenv
-                            system-lambda)
+                            source-name debug-name nil post-binding-lexenv)
     ;; ir1-convert-hairy-args can throw 'locall-already-let-converted
     ;; push optional-dispatch into the current component only after it
     ;; normally returned
@@ -884,8 +882,7 @@
 
 ;;; Convert a LAMBDA form into a LAMBDA leaf or an OPTIONAL-DISPATCH leaf.
 (defun ir1-convert-lambda (form &key (source-name '.anonymous.)
-                                     debug-name maybe-add-debug-catch
-                                     system-lambda)
+                                     debug-name maybe-add-debug-catch)
   (unless (consp form)
     (compiler-error "A ~S was found when expecting a lambda expression:~%  ~S"
                     (type-of form)
@@ -899,8 +896,6 @@
     (compiler-error
      "The lambda expression has a missing or non-list lambda list:~%  ~S"
      form))
-  (when (and system-lambda maybe-add-debug-catch)
-    (bug "Both SYSTEM-LAMBDA and MAYBE-ADD-DEBUG-CATCH specified"))
   (unless (or debug-name (neq '.anonymous. source-name))
     (setf debug-name (name-lambdalike form)))
   (binding* (((forms decls doc) (parse-body (cddr form) t))
@@ -925,25 +920,21 @@
              (forms (if (eq result-type *wild-type*)
                         forms
                         `((the ,(type-specifier result-type) (progn ,@forms)))))
-             (*allow-instrumenting* (and (not system-lambda) *allow-instrumenting*))
-             (res (cond ((or (find-if #'lambda-var-arg-info vars) keyp)
-                         (ir1-convert-hairy-lambda forms vars keyp
-                                                   allow-other-keys
-                                                   aux-vars aux-vals
-                                                   :post-binding-lexenv post-binding-lexenv
-                                                   :source-name source-name
-                                                   :debug-name debug-name
-                                                   :system-lambda system-lambda))
-                        (t
-                         (ir1-convert-lambda-body forms
-                                                  (add-types-for-fixed-args source-name vars)
-                                                  :aux-vars aux-vars
-                                                  :aux-vals aux-vals
-                                                  :post-binding-lexenv post-binding-lexenv
-                                                  :source-name source-name
-                                                  :debug-name debug-name
-                                                  :system-lambda system-lambda
-                                                  :local-policy local-policy)))))
+             (res (if (or (find-if #'lambda-var-arg-info vars) keyp)
+                      (ir1-convert-hairy-lambda forms vars keyp
+                                                allow-other-keys
+                                                aux-vars aux-vals
+                                                :post-binding-lexenv post-binding-lexenv
+                                                :source-name source-name
+                                                :debug-name debug-name)
+                      (ir1-convert-lambda-body forms
+                                               (add-types-for-fixed-args source-name vars)
+                                               :aux-vars aux-vars
+                                               :aux-vals aux-vals
+                                               :post-binding-lexenv post-binding-lexenv
+                                               :source-name source-name
+                                               :debug-name debug-name
+                                               :local-policy local-policy))))
     (when explicit-check
       (setf (getf (functional-plist res) 'explicit-check) explicit-check))
     (setf (functional-inline-expansion res) (or source-form form))
@@ -991,6 +982,14 @@
              (progn
                ,@forms))))))))
 
+;; FIXME: really should be an aspect of the lexical environment,
+;; but LEXENVs don't know whether they are toplevel or not.
+(defun has-toplevelness-decl (lambda-expr)
+  (dolist (expr (cddr lambda-expr)) ; Skip over (LAMBDA (ARGS))
+    (cond ((equal expr '(declare (top-level-form))) (return t))
+          ((typep expr '(or (cons (eql declare)) string))) ; DECL | DOCSTRING
+          (t (return nil)))))
+
 ;;; helper for LAMBDA-like things, to massage them into a form
 ;;; suitable for IR1-CONVERT-LAMBDA.
 (defun ir1-convert-lambdalike (thing
@@ -1026,6 +1025,8 @@
                  (info (info :function :info name)))
              (setf (functional-inlinep res) (info :function :inlinep name)
                    (defined-fun-same-block-p defined-fun-res) t)
+             (when (has-toplevelness-decl lambda-expression)
+               (setf (functional-top-level-defun-p res) t))
              ;; FIXME: Should non-entry block compiled defuns have
              ;; this propagate?
              (unless (and info
@@ -1035,7 +1036,7 @@
              ;; backward references to this function from following
              ;; top-level forms.
              (when simple-lexenv-p
-               (push res (defined-fun-functionals defined-fun-res)))
+               (setf (defined-fun-functional defined-fun-res) res))
              (unless (or
                       (eq (defined-fun-inlinep defined-fun-res) 'notinline)
                       ;; Don't treat recursive stubs like CAR as self-calls
@@ -1191,7 +1192,6 @@
                                   &key
                                   (source-name '.anonymous.)
                                   debug-name
-                                  system-lambda
                                   (policy (lexenv-policy *lexenv*)))
   (when (and (not debug-name) (eq '.anonymous. source-name))
     (setf debug-name (name-lambdalike fun)))
@@ -1221,8 +1221,7 @@
          (*inlining* (1+ *inlining*))
          (clambda (ir1-convert-lambda body
                                       :source-name source-name
-                                      :debug-name debug-name
-                                      :system-lambda system-lambda)))
+                                      :debug-name debug-name)))
     (setf (functional-inline-expanded clambda) t)
     clambda))
 
@@ -1261,7 +1260,7 @@
              (setf (gethash name free-funs) res)))
           ;; If FREE-FUNS has a previously converted definition
           ;; for this name, then blow it away and try again.
-          ((defined-fun-functionals found)
+          ((defined-fun-functional found)
            (remhash name free-funs)
            (get-defined-fun name lambda-list))
           (t found))))
@@ -1321,10 +1320,7 @@
   (unless (eq inlinep 'inline)
     (setf (defined-fun-inline-expansion var) nil))
   (let ((fun (ir1-convert-inline-lambda expansion
-                                        :source-name name
-                                        ;; prevent instrumentation of
-                                        ;; known function expansions
-                                        :system-lambda (and info t))))
+                                        :source-name name)))
     (setf (functional-inlinep fun) inlinep)
     (assert-new-definition var fun)
     (setf (defined-fun-inline-expansion var) expansion)
@@ -1361,7 +1357,7 @@
       ;; references to this function from following top-level
       ;; forms.
       (when (simple-lexical-environment-p *lexenv*)
-        (push fun (defined-fun-functionals var))))
+        (setf (defined-fun-functional var) fun)))
     fun))
 
 
@@ -1448,25 +1444,29 @@ is potentially harmful to any already-compiled callers using (SAFETY 0)."
 ;;; since those can always be reconstructed from a defstruct description.
 (defun %compiler-defun (name compile-toplevel inline-lambda extra-info)
   (let ((defined-fun nil)) ; will be set below if we're in the compiler
-    (when compile-toplevel
-      (with-single-package-locked-error
-          (:symbol name "defining ~S as a function")
-        (setf defined-fun
-              ;; Try to pass the lambda-list to GET-DEFINED-FUN if we can.
-              (if (atom inline-lambda)
-                  (get-defined-fun name)
-                  (get-defined-fun
-                   name (ecase (car inline-lambda)
-                         (lambda-with-lexenv (third inline-lambda))
-                         (lambda (second inline-lambda)))))))
-      (when (boundp '*lexenv*)
-        (aver (producing-fasl-file))
-        (if (member name (fun-names-in-this-file *compilation*) :test #'equal)
-            (warn 'duplicate-definition :name name)
-            (push name (fun-names-in-this-file *compilation*))))
-      ;; I don't know why this is guarded by (WHEN compile-toplevel),
-      ;; because regular old %DEFUN is going to call this anyway.
-      (%set-inline-expansion name defined-fun inline-lambda extra-info))
+    (cond (compile-toplevel
+           (with-single-package-locked-error
+               (:symbol name "defining ~S as a function")
+             (setf defined-fun
+                   ;; Try to pass the lambda-list to GET-DEFINED-FUN if we can.
+                   (if (atom inline-lambda)
+                       (get-defined-fun name)
+                       (get-defined-fun
+                        name (ecase (car inline-lambda)
+                               (lambda-with-lexenv (third inline-lambda))
+                               (lambda (second inline-lambda)))))))
+           (when (boundp '*lexenv*)
+             (aver (producing-fasl-file))
+             (if (member name (fun-names-in-this-file *compilation*) :test #'equal)
+                 (warn 'duplicate-definition :name name)
+                 (push name (fun-names-in-this-file *compilation*))))
+           ;; I don't know why this is guarded by (WHEN compile-toplevel),
+           ;; because regular old %DEFUN is going to call this anyway.
+           (%set-inline-expansion name defined-fun inline-lambda extra-info))
+          ((boundp 'sb-fasl::*current-fasl-group*)
+           (if (member name (sb-fasl::fasl-group-fun-names sb-fasl::*current-fasl-group*) :test #'equal)
+               (warn 'duplicate-definition :name name)
+               (push name (sb-fasl::fasl-group-fun-names sb-fasl::*current-fasl-group*)))))
 
     (become-defined-fun-name name)
 
