@@ -44,7 +44,9 @@
   (loop with policy = (lexenv-policy (node-lexenv call))
         for args on (basic-combination-args call)
         for var in (lambda-vars fun)
-        for name = (lambda-var-%source-name var)
+        for name = (or (and (lambda-var-arg-info var )
+                            (arg-info-key (lambda-var-arg-info var )))
+                       (lambda-var-%source-name var))
         do (assert-lvar-type (car args) (leaf-type var) policy
                              (if (eq (functional-kind fun) :optional)
                                  (make-local-call-context fun name)
@@ -56,35 +58,49 @@
 
 ;;; Given a local call CALL to FUN, find the associated argument LVARs
 ;;; of CALL corresponding to declared dynamic extent LAMBDA-VARs and
-;;; note them as dynamic extent LVARs. This operation is transitive,
-;;; because dynamic extent is contagious. In particular, the arguments
-;;; of any COMBINATIONs returning a stack-allocatable object in a
-;;; dynamic extent LVAR are dynamic extent as well if the argument
-;;; LVARs contain otherwise-inaccessible stack-allocatable subobjects
-;;; themselves.
-(defun recognize-dynamic-extent-lvars (call fun)
+;;; mark them as dynamic extent, setting up the cleanup corresponding
+;;; to the dynamic extent as well. We mark them now so that
+;;; optimizations optimizing away LVARs have the chance to propagate
+;;; the dynamic extent information. Environment analysis is
+;;; responsible for actually deciding if the lvars can be
+;;; dynamic-extent allocated, dealing with transitively marking the
+;;; otherwise-inaccessible parts of these values as dynamic extent as
+;;; well. This is because environment analysis happens after qll major
+;;; changes to the dataflow in IR1 have been done and it is clear
+;;; whether an LVAR is actually used by a combination which can
+;;; dynamic-extent allocate.
+(defun recognize-potentially-dynamic-extent-lvars (call fun)
   (declare (type combination call) (type clambda fun))
   ;; The block may end up being deleted due to cast optimization
   ;; caused by USE-GOOD-FOR-DX-P
   (unless (node-to-be-deleted-p call)
-    (let* ((*dx-lexenv* (node-lexenv call))
+    (let* (no-notes
            (dx-lvars
              (loop for arg in (basic-combination-args call)
                    for var in (lambda-vars fun)
                    for dx = (leaf-dynamic-extent var)
                    when (and dx arg (not (lvar-dynamic-extent arg)))
-                   append (handle-nested-dynamic-extent-lvars dx arg))))
+                   collect (cons dx arg)
+                   do
+                   (when (eq dx 'dynamic-extent-no-note)
+                     (setf no-notes dx)))))
       (when dx-lvars
         (let* ((entry (with-ir1-environment-from-node call
                         (make-entry)))
                (cleanup (make-cleanup :kind :dynamic-extent
                                       :mess-up entry
-                                      :nlx-info dx-lvars)))
+                                      :nlx-info dx-lvars
+                                      :dx-kind no-notes)))
           (setf (entry-cleanup entry) cleanup)
           (insert-node-before call entry)
           (setf (node-lexenv call)
                 (make-lexenv :default (node-lexenv call)
                              :cleanup cleanup))
+          (setf (ctran-next (node-prev call)) nil)
+          (let ((ctran (make-ctran)))
+            (with-ir1-environment-from-node call
+              (ir1-convert (node-prev call) ctran nil '(%cleanup-point))
+              (link-node-to-previous-ctran call ctran)))
           ;; Make CALL end its block, so that we have a place to
           ;; insert cleanup code.
           (node-ends-block call)
@@ -139,7 +155,7 @@
   (propagate-to-args call fun)
   (setf (basic-combination-kind call) :local)
   (sset-adjoin fun (lambda-calls-or-closes (node-home-lambda call)))
-  (recognize-dynamic-extent-lvars call fun)
+  (recognize-potentially-dynamic-extent-lvars call fun)
   (merge-tail-sets call fun)
   (change-ref-leaf ref fun)
   (values))
@@ -176,12 +192,20 @@
   (declare (type functional fun))
   (etypecase fun
     (clambda
-     (let* ((n-supplied (sb-xc:gensym))
+     (let* ((n-supplied (gensym))
             (nargs (length (lambda-vars fun)))
-            (temps (make-gensym-list nargs)))
+            (temps (make-gensym-list nargs))
+            (info (info :function :info (functional-%source-name fun)))
+            (types (and info
+                        (ir1-attributep (fun-info-attributes info) fixed-args)
+                        (loop for var in (lambda-vars fun)
+                              for temp in temps
+                              collect `(type ,(type-specifier (lambda-var-type var)) ,temp)))))
+
        `(lambda (,n-supplied ,@temps)
           (declare (type index ,n-supplied)
-                   (ignore ,n-supplied))
+                   (ignore ,n-supplied)
+                   ,@types)
           (%funcall ,fun ,@temps))))
     (optional-dispatch
      ;; Force conversion of all entries
@@ -189,7 +213,7 @@
      (let* ((min (optional-dispatch-min-args fun))
             (max (optional-dispatch-max-args fun))
             (more (optional-dispatch-more-entry fun))
-            (n-supplied (sb-xc:gensym))
+            (n-supplied (gensym))
             (temps (make-gensym-list max))
             (main (optional-dispatch-main-entry fun))
             (optional-vars (nthcdr min (lambda-vars main)))
@@ -288,8 +312,7 @@
                          *lexenv*))
            (xep (ir1-convert-lambda (make-xep-lambda-expression fun)
                                     :debug-name (debug-name
-                                                 'xep (leaf-debug-name fun))
-                                    :system-lambda t)))
+                                                 'xep (leaf-debug-name fun)))))
       (setf (functional-kind xep) :external
             (leaf-ever-used xep) t
             (functional-entry-fun xep) fun
@@ -422,7 +445,6 @@
            (clean-component component))))
      (unless did-something
        (return))))
-  (initial-eliminate-dead-code clambdas)
   (values))
 
 ;;; If policy is auspicious and CALL is not in an XEP and we don't seem
@@ -442,7 +464,18 @@
               (with-ir1-environment-from-node call
                 (let* ((*inline-expansions*
                          (register-inline-expansion original-functional call))
-                       (*lexenv* (functional-lexenv original-functional)))
+                       (functional-lexenv (functional-lexenv original-functional))
+                       (call-policy (lexenv-policy (node-lexenv call)))
+                       ;; The inline expansion should be converted
+                       ;; with the policy in effect at this call site,
+                       ;; not the policy saved in the original
+                       ;; functional.
+                       (*lexenv*
+                         (if (eq (lexenv-policy functional-lexenv)
+                                  call-policy)
+                             functional-lexenv
+                             (make-lexenv :default functional-lexenv
+                                          :policy call-policy))))
                   (values nil
                           (ir1-convert-lambda
                            (functional-inline-expansion original-functional)
@@ -689,8 +722,7 @@
                (%funcall ,entry ,@args))
             :debug-name (debug-name 'hairy-function-entry
                                     (lvar-fun-debug-name
-                                     (basic-combination-fun call)))
-            :system-lambda t))))
+                                     (basic-combination-fun call)))))))
     (convert-call ref call new-fun)
     (dolist (ref (leaf-refs entry))
       (convert-call-if-possible ref (lvar-dest (node-lvar ref))))))
@@ -1279,8 +1311,7 @@
 (defun maybe-let-convert (clambda &optional component)
   (declare (type clambda clambda)
            (type (or null component) component))
-  (unless (or (declarations-suppress-let-conversion-p clambda)
-              (functional-has-external-references-p clambda))
+  (unless (declarations-suppress-let-conversion-p clambda)
     ;; We only convert to a LET when the function is a normal local
     ;; function, has no XEP, and is referenced in exactly one local
     ;; call. Conversion is also inhibited if the only reference is in
@@ -1338,7 +1369,7 @@
   (declare (type cblock block1 block2))
   (or (eq block1 block2)
       (let ((cleanup2 (block-start-cleanup block2)))
-        (do-nested-cleanups (cleanup (block-end-lexenv block1) t)
+        (do-nested-cleanups (cleanup block1 t)
           (when (eq cleanup cleanup2)
             (return t))
           (case (cleanup-kind cleanup)
@@ -1408,7 +1439,6 @@
   (declare (type clambda fun))
   (when (and (not (functional-kind fun))
              (not (functional-entry-fun fun))
-             (not (functional-has-external-references-p fun))
              ;; If a functional is explicitly inlined, we don't want
              ;; to assignment convert it, as more call-site
              ;; specialization can be done with inlining.
