@@ -190,13 +190,26 @@
              `(progn
                 (define-vop (,(symbolicate 'fast- translate '-c/fixnum=>fixnum) fast-fixnum-binop-c)
                   (:translate ,translate)
-                  (:generator 1 (generate-fast-+-c r x (fixnumize (,translate y)))))
+                  (:generator 1
+                    (cond ((and (eq ',translate '-)
+                                (= y most-negative-fixnum))
+                           (inst lr temp-reg-tn (fixnumize y))
+                           (inst sub r x temp-reg-tn))
+                          (t
+                           (generate-fast-+-c r x (fixnumize (,translate y)))))))
                 (define-vop (,(symbolicate 'fast- translate '-c/signed=>signed) fast-signed-binop-c)
                   (:translate ,translate)
-                  (:generator ,untagged-penalty (generate-fast-+-c r x (,translate y))))
+                  (:generator ,untagged-penalty
+                    (cond ((and (eq ',translate '-)
+                                (= y (- (expt 2 (1- n-word-bits)))))
+                           (inst lr temp-reg-tn y)
+                           (inst sub r x temp-reg-tn))
+                          (t
+                           (generate-fast-+-c r x (,translate y))))))
                 (define-vop (,(symbolicate 'fast- translate '-c/unsigned=>unsigned) fast-unsigned-binop-c)
                   (:translate ,translate)
-                  (:generator ,untagged-penalty (generate-fast-+-c r x (,translate y)))))))
+                  (:generator ,untagged-penalty
+                    (generate-fast-+-c r x (,translate y)))))))
   (define-const-binop + 4)
   (define-const-binop - 4))
 
@@ -300,7 +313,7 @@
       (inst cmpdi amount 0)
       (inst neg ndesc amount)
       (inst bge positive)
-      (inst cmpdi ndesc 63)
+      (inst cmpldi ndesc 63)
       (inst srd result number ndesc)
       (inst ble done)
       (inst li result 0)
@@ -347,7 +360,7 @@
          (inst cmpdi amount 0)
          (inst neg ndesc amount)
          (inst bge positive)
-         (inst cmpdi ndesc 63)
+         (inst cmpldi ndesc 63)
          (inst srad result number ndesc)
          (inst ble done)
          ;; smear the sign bit into all bits
@@ -497,7 +510,7 @@
              fast-ash-left/unsigned=>unsigned))
 (deftransform ash-left-mod64 ((integer count)
                               ((unsigned-byte 64) (unsigned-byte 6)))
-  (when (sb-c::constant-lvar-p count)
+  (when (sb-c:constant-lvar-p count)
     (sb-c::give-up-ir1-transform))
   '(%primitive fast-ash-left-mod64/unsigned=>unsigned integer count))
 
@@ -869,7 +882,7 @@
   (:results (value :scs (unsigned-reg)))
   (:result-types unsigned-num))
 
-(define-vop (bignum-set word-index-set-nr)
+(define-vop (bignum-set word-index-set)
   (:variant bignum-digits-offset other-pointer-lowtag)
   (:translate sb-bignum:%bignum-set)
   (:args (object :scs (descriptor-reg))
@@ -1092,6 +1105,35 @@
   (:translate sb-bignum:%ashl)
   (:generator 1
     (inst sld result digit count)))
+
+(define-vop ()
+  (:translate fastrem-32)
+  (:policy :fast-safe)
+  (:args (dividend :scs (unsigned-reg))
+         (c :scs (unsigned-reg))
+         (divisor :scs (unsigned-reg)))
+  (:arg-types unsigned-num unsigned-num unsigned-num)
+  (:results (remainder :scs (unsigned-reg)))
+  (:result-types unsigned-num)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 10
+    (inst mulld temp dividend c)
+    (inst rldicl temp temp 0 32) ; drop the high 32 bits, keep the low 32 bits
+    (inst mulld temp temp divisor)
+    (inst srdi remainder temp 32))) ; take the high 32 bits
+(define-vop ()
+  (:translate fastrem-64)
+  (:policy :fast-safe)
+  (:args (dividend :scs (unsigned-reg))
+         (c :scs (unsigned-reg))
+         (divisor :scs (unsigned-reg)))
+  (:arg-types unsigned-num unsigned-num unsigned-num)
+  (:results (remainder :scs (unsigned-reg)))
+  (:result-types unsigned-num)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 10
+    (inst mulld temp dividend c) ; want only the low 64 bits
+    (inst mulhdu remainder temp divisor))) ; want only the high 64 bits
 
 (in-package "SB-C")
 
