@@ -173,11 +173,22 @@
   (:results (y :scs (descriptor-reg)))
   (:node-var node)
   (:note "float to pointer coercion")
+  #+gs-seg (:temporary (:sc unsigned-reg :offset 15) thread-tn)
   (:generator 13
-     (alloc-other y double-float-widetag double-float-size node)
+     (alloc-other double-float-widetag double-float-size y node nil thread-tn)
      (inst movsd (ea-for-df-desc y) x)))
 (define-move-vop move-from-double :move
   (double-reg) (descriptor-reg))
+(define-vop (!copy-dfloat)
+  (:args (x :scs (descriptor-reg) :to :save))
+  (:results (copy :scs (descriptor-reg)))
+  (:temporary (:sc unsigned-reg) bits)
+  (:node-var node)
+  #+gs-seg (:temporary (:sc unsigned-reg :offset 15) thread-tn)
+  (:generator 3
+     (alloc-other double-float-widetag double-float-size copy node nil thread-tn)
+     (loadw bits x double-float-value-slot other-pointer-lowtag)
+     (storew bits copy double-float-value-slot other-pointer-lowtag)))
 
 ;;; Move from a descriptor to a float register.
 (define-vop (move-to-single-reg)
@@ -225,10 +236,11 @@
 (define-vop (move-from-complex-single)
   (:args (x :scs (complex-single-reg) :to :save))
   (:results (y :scs (descriptor-reg)))
+  #+gs-seg (:temporary (:sc unsigned-reg :offset 15) thread-tn)
   (:node-var node)
   (:note "complex float to pointer coercion")
   (:generator 13
-     (alloc-other y complex-single-float-widetag complex-single-float-size node)
+     (alloc-other complex-single-float-widetag complex-single-float-size y node nil thread-tn)
      (inst movlps (ea-for-csf-data-desc y) x)))
 (define-move-vop move-from-complex-single :move
   (complex-single-reg) (descriptor-reg))
@@ -236,10 +248,11 @@
 (define-vop (move-from-complex-double)
   (:args (x :scs (complex-double-reg) :to :save))
   (:results (y :scs (descriptor-reg)))
+  #+gs-seg (:temporary (:sc unsigned-reg :offset 15) thread-tn)
   (:node-var node)
   (:note "complex float to pointer coercion")
   (:generator 13
-     (alloc-other y complex-double-float-widetag complex-double-float-size node)
+     (alloc-other complex-double-float-widetag complex-double-float-size y node nil thread-tn)
      (inst movapd (ea-for-cdf-data-desc y) x)))
 (define-move-vop move-from-complex-double :move
   (complex-double-reg) (descriptor-reg))
@@ -1070,7 +1083,7 @@
     complex-double-reg fp-complex-double-immediate complex-double-float
     movsd movapd cmppd movmskpd #b11))
 
-(macrolet ((define-</> (op single-name double-name &rest flags)
+(macrolet ((define (op single-name double-name &rest flags)
                `(progn
                   (define-vop (,double-name double-float-compare)
                     (:translate ,op)
@@ -1101,10 +1114,14 @@
                          (setf y (register-inline-constant (tn-value y))))
                         (t))
                       (inst comiss x y))))))
-  (define-</> < <single-float <double-float not :p :nc)
-  (define-</> > >single-float >double-float not :p :na)
-  (define-</> <= <=single-float <=double-float not :p :a)
-  (define-</> >= >=single-float >=double-float not :p :b))
+  ;;   UNORDERED:    ZF,PF,CF <- 111;
+  ;;   GREATER_THAN: ZF,PF,CF <- 000;
+  ;;   LESS_THAN:    ZF,PF,CF <- 001;
+  ;;   EQUAL:        ZF,PF,CF <- 100;
+  (define < <single-float <double-float not :p :nc)
+  (define > >single-float >double-float not :p :na)
+  (define <= <=single-float <=double-float not :p :a)
+  (define >= >=single-float >=double-float not :p :b))
 
 
 ;;;; conversion
@@ -1339,9 +1356,7 @@
        (double-stack
         (inst mov bits (ea (frame-byte-offset (tn-offset float)) rbp-tn)))
        (descriptor-reg
-        (inst mov bits (ea (- (ash double-float-value-slot word-shift)
-                              other-pointer-lowtag)
-                           float))))))
+        (loadw bits float double-float-value-slot other-pointer-lowtag)))))
 
 (define-vop (double-float-high-bits)
   (:args (float :scs (double-reg descriptor-reg)
@@ -1572,3 +1587,43 @@
   (:generator 2
      (move r x)
      (inst shufpd r r #b01)))
+
+#+round-float
+(progn
+  (define-vop ()
+    (:translate round-double)
+    (:policy :fast-safe)
+    (:args (x :scs (double-reg) :target r))
+    (:arg-types double-float (:constant symbol))
+    (:info mode)
+    (:results (r :scs (double-reg)))
+    (:result-types double-float)
+    (:generator 2
+      (unless (location= r x)
+        (inst xorpd r r))
+      (inst roundsd r x
+            (logior #b1000
+                    (ecase mode
+                      (:round 0)
+                      (:floor 1)
+                      (:ceiling 2)
+                      (:truncate 3))))))
+
+ (define-vop ()
+   (:translate round-single)
+   (:policy :fast-safe)
+   (:args (x :scs (single-reg) :target r))
+   (:arg-types single-float (:constant symbol))
+   (:info mode)
+   (:results (r :scs (single-reg)))
+   (:result-types single-float)
+   (:generator 2
+     (unless (location= r x)
+       (inst xorps r r))
+     (inst roundss r x
+           (logior #b1000
+                   (ecase mode
+                     (:round 0)
+                     (:floor 1)
+                     (:ceiling 2)
+                     (:truncate 3)))))))
